from metasim.utils import configclass

from .scene.scene_A import SCENE_A
from .scene.scene_B import SCENE_B
from .scene.scene_C import SCENE_C
from .scene.scene_D import SCENE_D
from .task.close_drawer_metacfg import CloseDrawerMetaCfg
from .task.lift_blue_block_drawer_metacfg import LiftBlueBlockDrawerMetaCfg
from .task.lift_blue_block_slider_metacfg import LiftBlueBlockSliderMetaCfg
from .task.lift_blue_block_table_metacfg import LiftBlueBlockTableMetaCfg
from .task.lift_pink_block_drawer_metacfg import LiftPinkBlockDrawerMetaCfg
from .task.lift_pink_block_slider_metacfg import LiftPinkBlockSliderMetaCfg
from .task.lift_pink_block_table_metacfg import LiftPinkBlockTableMetaCfg
from .task.lift_red_block_drawer_metacfg import LiftRedBlockDrawerMetaCfg
from .task.lift_red_block_slider_metacfg import LiftRedBlockSliderMetaCfg
from .task.lift_red_block_table_metacfg import LiftRedBlockTableMetaCfg
from .task.move_slider_left_metacfg import MoveSliderLeftMetaCfg
from .task.move_slider_right_metacfg import MoveSliderRightMetaCfg
from .task.open_drawer_metacfg import OpenDrawerMetaCfg
from .task.place_in_drawer_metacfg import PlaceInDrawerMetaCfg
from .task.place_in_slider_metacfg import PlaceInSliderMetaCfg
from .task.push_blue_block_left_metacfg import PushBlueBlockLeftMetaCfg
from .task.push_blue_block_right_metacfg import PushBlueBlockRightMetaCfg
from .task.push_into_drawer_metacfg import PushIntoDrawerMetaCfg
from .task.push_pink_block_left_metacfg import PushPinkBlockLeftMetaCfg
from .task.push_pink_block_right_metacfg import PushPinkBlockRightMetaCfg
from .task.push_red_block_left_metacfg import PushRedBlockLeftMetaCfg
from .task.push_red_block_right_metacfg import PushRedBlockRightMetaCfg
from .task.rotate_blue_block_left_metacfg import RotateBlueBlockLeftMetaCfg
from .task.rotate_blue_block_right_metacfg import RotateBlueBlockRightMetaCfg
from .task.rotate_pink_block_left_metacfg import RotatePinkBlockLeftMetaCfg
from .task.rotate_pink_block_right_metacfg import RotatePinkBlockRightMetaCfg
from .task.rotate_red_block_left_metacfg import RotateRedBlockLeftMetaCfg
from .task.rotate_red_block_right_metacfg import RotateRedBlockRightMetaCfg
from .task.stack_block_metacfg import StackBlockMetaCfg
from .task.unstack_block_metacfg import UnstackBlockMetaCfg


@configclass
class LiftRedBlockTableAMetaCfg(LiftRedBlockTableMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_red_block_table_a/v2"


@configclass
class LiftRedBlockTableBMetaCfg(LiftRedBlockTableMetaCfg):
    objects = SCENE_B.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_red_block_table_a/v2"  # TODO: use scene_a for testing


@configclass
class LiftRedBlockTableCMetaCfg(LiftRedBlockTableMetaCfg):
    objects = SCENE_C.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_red_block_table_a/v2"  # TODO: use scene_a for testing


@configclass
class LiftRedBlockTableDMetaCfg(LiftRedBlockTableMetaCfg):
    objects = SCENE_D.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_red_block_table_a/v2"  # TODO: use scene_a for testing


@configclass
class LiftRedBlockSliderAMetaCfg(LiftRedBlockSliderMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_red_block_slider_a/v2"


@configclass
class LiftRedBlockDrawerAMetaCfg(LiftRedBlockDrawerMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_red_block_drawer_a/v2"


@configclass
class LiftBlueBlockTableAMetaCfg(LiftBlueBlockTableMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_blue_block_table_a/v2"


@configclass
class LiftBlueBlockSliderAMetaCfg(LiftBlueBlockSliderMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_blue_block_slider_a/v2"


@configclass
class LiftBlueBlockDrawerAMetaCfg(LiftBlueBlockDrawerMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_blue_block_drawer_a/v2"


@configclass
class LiftPinkBlockTableAMetaCfg(LiftPinkBlockTableMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_pink_block_table_a/v2"


@configclass
class LiftPinkBlockSliderAMetaCfg(LiftPinkBlockSliderMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_pink_block_slider_a/v2"


@configclass
class LiftPinkBlockDrawerAMetaCfg(LiftPinkBlockDrawerMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/lift_pink_block_drawer_a/v2"


@configclass
class PlaceInSliderAMetaCfg(PlaceInSliderMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/place_in_slider_a/v2"


@configclass
class PlaceInDrawerAMetaCfg(PlaceInDrawerMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/place_in_drawer_a/v2"


@configclass
class RotateRedBlockRightAMetaCfg(RotateRedBlockRightMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/rotate_red_block_right_a/v2"


@configclass
class RotateRedBlockLeftAMetaCfg(RotateRedBlockLeftMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/rotate_red_block_left_a/v2"


@configclass
class RotateBlueBlockRightAMetaCfg(RotateBlueBlockRightMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/rotate_blue_block_right_a/v2"


@configclass
class RotateBlueBlockLeftAMetaCfg(RotateBlueBlockLeftMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/rotate_blue_block_left_a/v2"


@configclass
class RotatePinkBlockRightAMetaCfg(RotatePinkBlockRightMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/rotate_pink_block_right_a/v2"


@configclass
class RotatePinkBlockLeftAMetaCfg(RotatePinkBlockLeftMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/rotate_pink_block_left_a/v2"


@configclass
class PushRedBlockRightAMetaCfg(PushRedBlockRightMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/push_red_block_right_a/v2"


@configclass
class PushRedBlockLeftAMetaCfg(PushRedBlockLeftMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/push_red_block_left_a/v2"


@configclass
class PushBlueBlockRightAMetaCfg(PushBlueBlockRightMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/push_blue_block_right_a/v2"


@configclass
class PushBlueBlockLeftAMetaCfg(PushBlueBlockLeftMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/push_blue_block_left_a/v2"


@configclass
class PushPinkBlockRightAMetaCfg(PushPinkBlockRightMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/push_pink_block_right_a/v2"


@configclass
class PushPinkBlockLeftAMetaCfg(PushPinkBlockLeftMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/push_pink_block_left_a/v2"


@configclass
class MoveSliderLeftAMetaCfg(MoveSliderLeftMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/move_slider_left_a/v2"


@configclass
class MoveSliderRightAMetaCfg(MoveSliderRightMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/move_slider_right_a/v2"


@configclass
class OpenDrawerAMetaCfg(OpenDrawerMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/open_drawer_a/v2"


@configclass
class CloseDrawerAMetaCfg(CloseDrawerMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/close_drawer_a/v2"


@configclass
class StackBlockAMetaCfg(StackBlockMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/stack_block_a/v2"


@configclass
class UnstackBlockAMetaCfg(UnstackBlockMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/unstack_block_a/v2"


@configclass
class PushIntoDrawerAMetaCfg(PushIntoDrawerMetaCfg):
    objects = SCENE_A.objects
    traj_filepath = "roboverse_data/trajs/calvin/push_into_drawer_a/v2"



## XXX:
## 1. Currently we use global variables to track the progress, which is not a good practice.
## TODO:
## 1. Check the missing demos first, then collect the missing part? In this way, there won't be any global variables
## 2. Or, combine tot_success, tot_give_up, global_step and pbar into a seperate class, maybe called ProgressManager. In this way, there won't be any global variables

from __future__ import annotations

#########################################
## Setup logging
#########################################
from loguru import logger as log
from rich.logging import RichHandler

log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])


#########################################
### Add command line arguments
#########################################
import math
from dataclasses import dataclass
from typing import Literal

import rootutils
import tyro

rootutils.setup_root(__file__, pythonpath=True)
from metasim.cfg.randomization import RandomizationMetaCfg
from metasim.cfg.render import RenderMetaCfg


@dataclass
class Args:
    random: RandomizationMetaCfg
    """Domain randomization options"""
    render: RenderMetaCfg
    """Renderer options"""
    task: str
    """Task name"""
    robot: str = "franka"
    """Robot name"""
    num_envs: int = 1
    """Number of parallel environments, find a proper number for best performance on your machine"""
    sim: Literal["isaaclab", "mujoco", "isaacgym"] = "isaaclab"
    """Simulator backend"""
    demo_start_idx: int | None = None
    """The index of the first demo to collect, None for all demos"""
    max_demo_idx: int | None = None
    """Maximum number of demos to collect, None for all demos"""
    retry_num: int = 0
    """Number of retries for a failed demo"""
    headless: bool = False
    """Run in headless mode"""
    table: bool = True
    """Try to add a table"""
    tot_steps_after_success: int = 20
    """Maximum number of steps to collect after success, or until run out of demo"""
    all_rerender: bool = False
    """Rerender all traj, even if some are already rendered"""
    split: Literal["train", "val", "test", "all"] = "all"
    """Split to collect"""
    cust_name: str | None = None
    """Custom name for the dataset"""
    scene: str | None = None
    """Scene name"""

    def __post_init__(self):
        if self.random.table and not self.table:
            log.warning("Cannot enable table randomization without a table, disabling table randomization")
            self.random.table = False

        if self.max_demo_idx is None:
            self.max_demo_idx = math.inf

        if self.demo_start_idx is None:
            self.demo_start_idx = 0

        log.info(f"Args: {self}")


args = tyro.cli(Args)


#########################################
### Import packages
#########################################
if args.sim == "isaacgym":
    import isaacgym  # noqa: F401
    from isaacgym import gymapi, gymtorch, gymutil  # noqa: F401

import multiprocessing as mp
import os

from tqdm.rich import tqdm_rich as tqdm

from metasim.cfg.cameras import PinholeCameraMetaCfg
from metasim.cfg.scenario import ScenarioMetaCfg


###########################################################
## Utils
###########################################################
def get_actions(all_actions, handler, demo_idxs: list[int]):
    from metasim.sim.base import BaseSimHandler

    assert isinstance(handler, BaseSimHandler)

    action_idxs = handler.episode_length_buf
    actions = [
        all_actions[demo_idx][action_idx] if action_idx < len(all_actions[demo_idx]) else all_actions[demo_idx][-1]
        for demo_idx, action_idx in zip(demo_idxs, action_idxs)
    ]

    return actions


def get_states(all_states, handler, demo_idxs: list[int]):
    from metasim.sim.base import BaseSimHandler

    assert isinstance(handler, BaseSimHandler)
    action_idxs = handler.episode_length_buf
    states = [
        all_states[demo_idx][action_idx] if action_idx < len(all_states[demo_idx]) else all_states[demo_idx][-1]
        for demo_idx, action_idx in zip(demo_idxs, action_idxs)
    ]
    return states


def get_run_out(all_actions, handler, demo_idxs: list[int]) -> list[bool]:
    action_idxs = handler.episode_length_buf
    run_out = [action_idx >= len(all_actions[demo_idx]) for demo_idx, action_idx in zip(demo_idxs, action_idxs)]
    return run_out


def save_demo_mp(save_req_queue: mp.Queue):
    from metasim.utils.save_util import save_demo

    while (save_request := save_req_queue.get()) is not None:
        demo = save_request["demo"]
        save_dir = save_request["save_dir"]
        log.info(f"Received save request, saving to {save_dir}")
        save_demo(save_dir, demo)


###########################################################
## Global Variables
###########################################################
global global_step, tot_success, tot_give_up
tot_success = 0
tot_give_up = 0
global_step = 0


###########################################################
## Core Utils
###########################################################
class DemoCollector:
    def __init__(self, handler):
        from metasim.sim.base import BaseSimHandler

        assert isinstance(handler, BaseSimHandler)
        self.handler = handler
        self.cache: dict[int, list[dict]] = {}
        self.save_request_queue = mp.Queue()
        self.save_proc = mp.Process(target=save_demo_mp, args=(self.save_request_queue,))
        self.save_proc.start()

    def create(self, demo_idx: int):
        assert demo_idx not in self.cache
        assert isinstance(demo_idx, int)
        self.cache[demo_idx] = []

    def add(self, data_dict: dict | None, demo_idx: int):
        if data_dict is None:
            log.warning("Skipping adding obs to DemoCollector because obs is None")
        assert demo_idx in self.cache
        self.cache[demo_idx].append(data_dict)

    def save(self, demo_idx: int):
        assert demo_idx in self.cache
        TaskName = self.handler.task.__class__.__name__.replace("MetaCfg", "")
        if args.cust_name is not None:
            additional_str = "-" + str(args.cust_name)
        else:
            additional_str = ""
        save_dir = f"data_isaaclab/demo_{args.sim}/{TaskName}-Level{args.random.level}{additional_str}/robot-{args.robot}/demo_{demo_idx:04d}"
        os.makedirs(save_dir, exist_ok=True)
        log.info(f"Saving demo {demo_idx} to {save_dir}")

        ## Option 1: Save immediately, blocking and slower

        from metasim.utils.save_util import save_demo

        save_demo(save_dir, self.cache[demo_idx])

        ## Option 2: Save in a separate process, non-blocking, not friendly to KeyboardInterrupt, TODO: fix
        ## TODO: see https://isaac-sim.github.io/IsaacLab/main/source/refs/troubleshooting.html#preventing-memory-leaks-in-the-simulator
        # self.save_request_queue.put({"demo": self.cache[demo_idx], "save_dir": save_dir})

    def delete(self, demo_idx: int):
        assert demo_idx in self.cache
        del self.cache[demo_idx]

    def final(self):
        self.save_request_queue.put(None)  # signal to save_demo_mp to exit
        self.save_proc.join()
        assert self.cache == {}


class DemoIndexer:
    def __init__(
        self,
        save_root_dir: str,
        start_idx: int,
        end_idx: int,
        pbar: tqdm,
        overwrite: bool = True,
    ):
        self.save_root_dir = save_root_dir
        self._next_idx = start_idx
        self.end_idx = end_idx
        self.pbar = pbar
        self.overwrite = overwrite

    @property
    def next_idx(self):
        return self._next_idx

    def move_on(self):
        self._next_idx += 1
        if not self.overwrite:
            while os.path.exists(f"{self.save_root_dir}/demo_{self._next_idx:04d}/metadata.json"):
                global global_step, tot_success, tot_give_up
                tot_success += 1
                self.pbar.set_description(f"Frame {global_step} Success {tot_success} Giveup {tot_give_up}")
                self.pbar.update(1)
                log.info(f"Demo {self._next_idx} already exists, skipping...")
                self._next_idx += 1


###########################################################
## Main
###########################################################
def main():
    global global_step, tot_success, tot_give_up
    num_envs: int = args.num_envs

    # specificly for isaacgym
    if args.sim == "isaacgym":
        from isaacgym import gymapi, gymtorch, gymutil  # noqa: F401

    ## Import put here to support isaacgym

    import torch

    from metasim.constants import SimType
    from metasim.utils.demo_util import get_traj
    from metasim.utils.setup_util import get_robot, get_sim_env_class, get_task

    handler_class = get_sim_env_class(SimType(args.sim))
    task = get_task(args.task)
    robot = get_robot(args.robot)
    camera = PinholeCameraMetaCfg(
        name="camera",
        data_types=["rgb", "depth"],
        width=256,
        height=256,
        pos=(1.5, 0.0, 1.5),
        look_at=(0.0, 0.0, 0.0),
    )
    # realsense
    # import numpy as np
    # camera = PinholeCameraMetaCfg(
    #     name="camera",
    #     data_types=["rgb", "depth"],
    #     width=640,
    #     height=360,
    #     pos=(1.3, 0.0, 0.7),
    #     look_at=(0.0, 0.0, 0.0),
    #     focal_length=1.88,
    #     horizontal_aperture=float(2 * 1.88 * np.tan(71.28 / 180 * np.pi / 2)),
    # )
    scenario = ScenarioMetaCfg(
        task=task,
        robot=robot,
        scene=args.scene,
        cameras=[camera],
        random=args.random,
        try_add_table=args.table,
        render=args.render,
        split=args.split,
        sim=args.sim,
    )
    env = handler_class(scenario, num_envs, args.headless)

    ## Data
    assert os.path.exists(task.traj_filepath), f"Trajectory file does not exist: {task.traj_filepath}"
    init_states, all_actions, all_states = get_traj(task, robot, env.handler)

    tot_demo = len(all_actions)
    if args.split == "train":
        init_states = init_states[: int(tot_demo * 0.9)]
        all_actions = all_actions[: int(tot_demo * 0.9)]
        all_states = all_states[: int(tot_demo * 0.9)]
    elif args.split == "val" or args.split == "test":
        init_states = init_states[int(tot_demo * 0.9) :]
        all_actions = all_actions[int(tot_demo * 0.9) :]
        all_states = all_states[int(tot_demo * 0.9) :]

    n_demo = len(all_actions)
    log.info(f"Collecting from {args.split} split, {n_demo} out of {tot_demo} demos")

    ########################################################
    ## Main
    ########################################################
    if args.max_demo_idx > n_demo:
        log.warning(
            f"Max demo {args.max_demo_idx} is greater than the number of demos in the dataset {n_demo}, using {n_demo}"
        )
    max_demo = min(args.max_demo_idx, n_demo)
    try_num = args.retry_num + 1

    ###########################################################
    ##   State Machine Diagram
    ###########################################################
    ##   CollectingDemo --> Success: env success
    ##   CollectingDemo --> Timeout: env timeout or run_out
    ##
    ##   Success --> FinalizeCollectingDemo
    ##
    ##   FinalizeCollectingDemo --> NextDemo: run_out or steps_after_success >= args.tot_steps_after_success
    ##
    ##   Timeout --> CollectingDemo: failure_count < try_num
    ##   Timeout --> NextDemo: failure_count >= try_num
    ##
    ##   NextDemo --> CollectingDemo: next_demo_idx < max_demo
    ##   NextDemo --> Finished: next_demo_idx >= max_demo
    ##
    ##   All Finished --> Exit

    ## Setup
    collector = DemoCollector(env.handler)
    pbar = tqdm(total=max_demo - args.demo_start_idx, desc="Collecting demos")

    ## State variables
    failure_count = [0] * env.handler.num_envs
    steps_after_success = [0] * env.handler.num_envs
    finished = [False] * env.handler.num_envs
    TaskName = env.handler.task.__class__.__name__.replace("MetaCfg", "")

    if args.cust_name is not None:
        additional_str = "-" + str(args.cust_name)
    else:
        additional_str = ""
    demo_indexer = DemoIndexer(
        save_root_dir=f"data_isaaclab/demo_{args.sim}/{TaskName}-Level{args.random.level}{additional_str}/robot-{args.robot}",
        start_idx=0,
        end_idx=max_demo,
        pbar=pbar,
        overwrite=args.all_rerender,
    )
    demo_idxs = []
    for demo_idx in range(env.handler.num_envs):
        demo_idxs.append(demo_indexer.next_idx)
        demo_indexer.move_on()
    log.info(f"Initialize with demo idxs: {demo_idxs}")

    ## Reset before first step
    env.handler.set_states([init_states[demo_idx] for demo_idx in demo_idxs])
    obs, extras = env.handler.reset()

    ## Initialize
    for demo_idx in demo_idxs:
        collector.create(demo_idx)

    ## Main Loop
    while not all(finished):
        pbar.set_description(f"Frame {global_step} Success {tot_success} Giveup {tot_give_up}")
        actions = get_actions(all_actions, env.handler, demo_idxs)
        obs, reward, success, time_out, extras = env.step(actions)
        run_out = get_run_out(all_actions, env.handler, demo_idxs)

        for env_id in range(env.handler.num_envs):
            if finished[env_id]:
                continue

            demo_idx = demo_idxs[env_id]
            obs_cur_env = {k: v[env_id] for k, v in obs.items()} if obs is not None else None
            collector.add(obs_cur_env, demo_idx)

        for env_id in success.nonzero().squeeze(-1).tolist():
            if finished[env_id]:
                continue

            ## CollectingDemo --> Success
            demo_idx = demo_idxs[env_id]
            if steps_after_success[env_id] == 0:
                log.info(f"Demo {demo_idx} in Env {env_id} succeeded!")
                tot_success += 1
                pbar.update(1)
                pbar.set_description(f"Frame {global_step} Success {tot_success} Giveup {tot_give_up}")

            ## Success --> FinalizeCollectingDemo
            if not run_out[env_id] and steps_after_success[env_id] < args.tot_steps_after_success:
                steps_after_success[env_id] += 1
            else:
                ## FinalizeCollectingDemo --> NextDemo
                steps_after_success[env_id] = 0
                collector.save(demo_idx)
                collector.delete(demo_idx)

                if demo_indexer.next_idx < max_demo:
                    ## NextDemo --> CollectingDemo
                    demo_idxs[env_id] = demo_indexer.next_idx
                    env.handler.set_states([init_states[demo_idx] for demo_idx in demo_idxs], env_ids=[env_id])
                    env.handler.reset(env_ids=[env_id])
                    collector.create(demo_indexer.next_idx)
                    demo_indexer.move_on()
                    run_out[env_id] = False
                else:
                    ## NextDemo --> Finished
                    finished[env_id] = True

        for env_id in (time_out | torch.tensor(run_out, device=time_out.device)).nonzero().squeeze(-1).tolist():
            if finished[env_id]:
                continue

            ## CollectingDemo --> Timeout
            demo_idx = demo_idxs[env_id]
            log.info(f"Demo {demo_idx} in Env {env_id} timed out!")
            collector.delete(demo_idx)
            failure_count[env_id] += 1

            if failure_count[env_id] < try_num:
                ## Timeout --> CollectingDemo
                log.info(f"Demo {demo_idx} failed {failure_count[env_id]} times, retrying...")
                env.handler.set_states([init_states[demo_idx] for demo_idx in demo_idxs], env_ids=[env_id])
                env.handler.reset(env_ids=[env_id])
                collector.create(demo_idx)
            else:
                ## Timeout --> NextDemo
                log.error(f"Demo {demo_idx} failed too many times, giving up")
                failure_count[env_id] = 0
                tot_give_up += 1
                pbar.update(1)
                pbar.set_description(f"Frame {global_step} Success {tot_success} Giveup {tot_give_up}")

                if demo_indexer.next_idx < max_demo:
                    ## NextDemo --> CollectingDemo
                    demo_idxs[env_id] = demo_indexer.next_idx
                    env.handler.set_states([init_states[demo_idx] for demo_idx in demo_idxs], env_ids=[env_id])
                    env.handler.reset(env_ids=[env_id])
                    collector.create(demo_indexer.next_idx)
                    demo_indexer.move_on()
                else:
                    ## NextDemo --> Finished
                    finished[env_id] = True

        global_step += 1

    log.info("Finalizing")
    collector.final()
    env.close()


if __name__ == "__main__":
    main()



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class InsertOntoSquarePegMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 600
    traj_filepath = "roboverse_data/trajs/rlbench/insert_onto_square_peg/v2"
    objects = [
        RigidObjMetaCfg(
            name="square_ring",
            filepath="roboverse_data/assets/rlbench/insert_onto_square_peg/square_ring/usd/square_ring.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        PrimitiveCubeMetaCfg(
            name="pillar0",
            size=[0.025, 0.025, 0.12],
            physics=PhysicStateType.GEOM,
            color=[1.0, 0.0, 1.0],
        ),
        PrimitiveCubeMetaCfg(
            name="pillar1",
            physics=PhysicStateType.GEOM,
            size=[0.025, 0.025, 0.12],
            color=[0.5, 0.5, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="pillar2",
            physics=PhysicStateType.GEOM,
            size=[0.025, 0.025, 0.12],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="square_base",
            physics=PhysicStateType.GEOM,
            size=[0.4, 0.1, 0.02],
            color=[0.49, 0.38, 0.29],
        ),
    ]
    # TODO: add checker



from metasim.cfg.checkers import RunChecker
from metasim.utils import configclass

from .base_locomotion_env import BaseLocomotionReward, HumanoidTaskMetaCfg


class RunReward(BaseLocomotionReward):
    _move_speed = 5
    success_bar = 700


@configclass
class RunMetaCfg(HumanoidTaskMetaCfg):
    episode_length = 1000
    traj_filepath = "roboverse_data/trajs/humanoidbench/run/v2/initial_state_v2.json"
    checker = RunChecker()
    reward_functions = [RunReward]
    reward_weights = [1.0]



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class BlockPyramidMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/block_pyramid/v2"
    objects = [
        RigidObjMetaCfg(
            name="block_pyramid_plane",
            filepath="roboverse_data/assets/rlbench/block_pyramid/block_pyramid_plane/usd/block_pyramid_plane.usd",
            physics=PhysicStateType.GEOM,
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_block0",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[0.0, 0.5, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_block1",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[0.0, 0.5, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_block2",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[0.0, 0.5, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_block3",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[0.0, 0.5, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_block4",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[0.0, 0.5, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_block5",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[0.0, 0.5, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_distractor_block0",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_distractor_block1",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_distractor_block2",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_distractor_block3",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_distractor_block4",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="block_pyramid_distractor_block5",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.0375, 0.0375, 0.0375],
            color=[1.0, 0.0, 0.0],
        ),
    ]
    # TODO: add checker



import copy

import numpy as np
from loguru import logger as log

from metasim.cfg.cameras import PinholeCameraMetaCfg
from metasim.cfg.tasks import BaseTaskMetaCfg
from metasim.constants import BenchmarkType

_phi_theta_candidates = np.loadtxt("metasim/sim/isaaclab/cfg/randomization/camera_pos_candidates.txt")


def randomize_camera_pose(
    original_camera: PinholeCameraMetaCfg,
    obj_pos: tuple[float, float, float],
    robot_quat: tuple[float, float, float, float],
    mode: str,
    task: BaseTaskMetaCfg,
) -> PinholeCameraMetaCfg:
    ## XXX: We assume that when robot front side faces X+ in its local coordinate,

    randomized_camera = copy.deepcopy(original_camera)
    if mode == "semisphere":
        distance = np.random.uniform(2, 4)
        theta_to_robot = np.random.uniform(0, 2 * np.pi)
        phi = np.random.uniform(0, 70 / 180 * np.pi)
        look_at_offset = np.random.uniform([-0.2, -0.2, 0], [0.2, 0.2, 0])
    elif mode == "front":
        distance = 2.0
        theta_to_robot = 0.0
        phi = 60.0 / 180.0 * np.pi
        look_at_offset = np.array([0.0, 0.0, 0.0])
    elif mode == "front_uniform_random":
        distance = np.random.uniform(1.5, 2.0)
        theta_to_robot = np.random.normal(0, np.pi * (2 / 3))
        phi = np.random.uniform(70, 50) / 180.0 * np.pi
        look_at_offset = np.array([0.0, 0.0, 0.0])
    elif mode == "front_triangular_random":
        distance = np.random.uniform(1.5, 2.0)
        theta_to_robot = np.random.triangular(-np.pi * (2 / 3), 0, np.pi * (2 / 3))
        phi = np.random.uniform(70, 50) / 180.0 * np.pi
        look_at_offset = np.array([0.0, 0.0, 0.0])
    elif mode == "front_select":
        distance = 1.5
        phi, theta_to_robot = _phi_theta_candidates[np.random.randint(0, _phi_theta_candidates.shape[0])]
        look_at_offset = np.array([0.0, 0.0, 0.0])
        # XXX: Avoid table blocking the view
        if task.source_benchmark == BenchmarkType.CALVIN:
            distance = 2.0
            theta_to_robot = (theta_to_robot - np.pi / 2) / 240 * 180

    else:
        raise ValueError(f"Unknown mode: {mode}")

    robot_theta = np.arctan2(robot_quat[1], robot_quat[0])
    log.debug(f"robot_theta: {robot_theta / np.pi * 180:.2f}, theta_to_robot: {theta_to_robot / np.pi * 180:.2f}")
    theta = robot_theta + theta_to_robot

    pos = (np.array([np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)]) * distance).tolist()
    look_at = (np.array(obj_pos) + look_at_offset).tolist()
    randomized_camera.pos = pos
    randomized_camera.look_at = look_at
    randomized_camera.focus_distance = distance
    return randomized_camera



from metasim.cfg.checkers import DetectedChecker, OrOp, RelativeBboxDetector
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, TaskType
from metasim.utils import configclass

slider_detector = RelativeBboxDetector(
    base_obj_name="table",
    relative_pos=(-0.181, 0.062, 0.501),
    relative_quat=(1.0, 0.0, 0.0, 0.0),
    checker_lower=(-0.3, -0.075, -0.05),
    checker_upper=(0.3, 0.075, 0.05),
    # debug_vis=True,
)


@configclass
class PlaceInSliderMetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.CALVIN
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    checker = OrOp([
        DetectedChecker(
            obj_name="block_pink",
            detector=slider_detector,
            ignore_if_first_check_success=True,
        ),
        DetectedChecker(
            obj_name="block_blue",
            detector=slider_detector,
            ignore_if_first_check_success=True,
        ),
        DetectedChecker(
            obj_name="block_red",
            detector=slider_detector,
            ignore_if_first_check_success=True,
        ),
    ])



## TODO:
## 1. The task implementation is not exactly same as the source benchmark
##    In this implementation, the hole on the base is slightly enlarged to reduce the difficulty due to the dynamics gap


from metasim.cfg.checkers.checkers import DetectedChecker
from metasim.cfg.checkers.detectors import RelativeBboxDetector
from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .maniskill_task_metacfg import ManiskillTaskMetaCfg


@configclass
class PlugChargerMetaCfg(ManiskillTaskMetaCfg):
    episode_length = 250
    objects = [
        RigidObjMetaCfg(
            name="base",
            filepath="roboverse_data/assets/maniskill/charger/base/base.usd",
            urdf_path="roboverse_data/assets/maniskill/charger/base.urdf",
            physics=PhysicStateType.GEOM,
            fix_base_link=True,
        ),
        RigidObjMetaCfg(
            name="charger",
            filepath="roboverse_data/assets/maniskill/charger/charger/charger.usd",
            urdf_path="roboverse_data/assets/maniskill/charger/charger.urdf",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "roboverse_data/trajs/maniskill/plug_charger/trajectory-franka_v2.pkl"

    checker = DetectedChecker(
        obj_name="charger",
        detector=RelativeBboxDetector(
            base_obj_name="base",
            relative_quat=[1, 0, 0, 0],
            relative_pos=[0, 0, 0],
            checker_lower=[-0.02, -0.075, -0.075],
            checker_upper=[0.02, 0.075, 0.075],
            # debug_vis=True,
        ),
    )



import torch

from metasim.constants import TaskType
from metasim.utils import configclass

from ..base_task_metacfg import BaseTaskMetaCfg


@configclass
class AntIsaacGymMetaCfg(BaseTaskMetaCfg):
    episode_length = 100
    objects = []
    traj_filepath = None
    task_type = TaskType.LOCOMOTION

    def reward_fn(self, states):
        # Reward constants
        up_weight = 0.1
        heading_weight = 0.5
        actions_cost_scale = 0.005
        energy_cost_scale = 0.001
        joints_at_limit_cost_scale = 0.1
        termination_height = 0.3
        death_cost = -1.0

        # Handle both multi-env (IsaacGym) and single-env (Mujoco) formats
        rewards = []
        for env_state in states:
            # Get ant states
            ant_state = env_state["ant"]

            # Extract positions, rotations and velocities
            pos = ant_state["pos"]
            rot = ant_state["rot"]  # quaternion [w, x, y, z]
            vel = ant_state["vel"]
            ang_vel = ant_state["ang_vel"]

            # Joint states
            joint_pos = torch.tensor([v for v in ant_state["dof_pos"].values()])
            joint_vel = torch.tensor([v for v in ant_state["dof_vel"].values()])

            # 1. Calculate up alignment reward (z-axis alignment with world up)
            # Convert quaternion to get body's up vector
            # For a quaternion [w, x, y, z], the up vector transform is:
            # up_x = 2 * (x*z + w*y)
            # up_y = 2 * (y*z - w*x)
            # up_z = 1 - 2 * (x*x + y*y)
            if len(rot) == 4:  # quaternion
                up_z = 1.0 - 2.0 * (rot[1] ** 2 + rot[2] ** 2)
            else:  # rotation matrix or other format
                # Assuming the third row of rotation matrix represents up vector
                up_z = rot[2]

            up_reward = torch.zeros(1, dtype=torch.float32)
            if up_z > 0.93:  # ant is upright
                up_reward = torch.tensor([up_weight], dtype=torch.float32)

            # 2. Heading reward (moving forward)
            # Use x-velocity as forward motion metric
            vel_x = vel[0]
            heading_reward = heading_weight * torch.clamp(vel_x / 0.8, 0.0, 1.0)

            # 3. Action costs
            # Since we don't have the actions here, use joint velocities as proxy
            actions_cost = torch.sum(joint_vel**2)

            # 4. Electricity costs (joint velocities * positions)
            electricity_cost = torch.sum(torch.abs(joint_vel * joint_pos))

            # 5. Joint limit penalties
            joint_names = list(ant_state["dof_pos"].keys())
            joint_limits = {
                "hip_1": (-0.6981, 0.6981),
                "ankle_1": (0.5236, 1.7453),
                "hip_2": (-0.6981, 0.6981),
                "ankle_2": (-1.7453, -0.5236),
                "hip_3": (-0.6981, 0.6981),
                "ankle_3": (-1.7453, -0.5236),
                "hip_4": (-0.6981, 0.6981),
                "ankle_4": (0.5236, 1.7453),
            }

            joints_at_limit = torch.tensor([
                float(
                    abs(ant_state["dof_pos"][j] - joint_limits[j][0]) < 0.01
                    or abs(ant_state["dof_pos"][j] - joint_limits[j][1]) < 0.01
                )
                for j in joint_names
            ])
            dof_at_limit_cost = torch.sum(joints_at_limit)

            # 6. Alive reward (constant reward for being alive)
            alive_reward = torch.tensor([0.5], dtype=torch.float32)

            # Height termination
            height = pos[2]
            total_reward = (
                alive_reward
                + up_reward
                + heading_reward
                - actions_cost_scale * actions_cost
                - energy_cost_scale * electricity_cost
                - joints_at_limit_cost_scale * dof_at_limit_cost
            )

            # Death penalty
            if height < termination_height:
                total_reward = torch.tensor([death_cost], dtype=torch.float32)

            rewards.append(total_reward)

        # Sum individual rewards for each environment
        return torch.cat(rewards)



from __future__ import annotations

from metasim.utils import configclass

from .base_robot_metacfg import BaseActuatorMetaCfg, BaseRobotMetaCfg


@configclass
class H1SimpleHandMetaCfg(BaseRobotMetaCfg):
    name: str = "h1_simple_hand"
    num_joints: int = 52
    mjcf_path: str = "roboverse_data/robots/h1_simple_hand/mjcf/h1_simple_hand.xml"

    enabled_gravity: bool = True
    freejoint: bool = True
    enabled_self_collisions: bool = False

    actuators: dict[str, BaseActuatorMetaCfg] = {
        "left_hip_yaw": BaseActuatorMetaCfg(),
        "left_hip_roll": BaseActuatorMetaCfg(),
        "left_hip_pitch": BaseActuatorMetaCfg(),
        "left_knee": BaseActuatorMetaCfg(),
        "left_ankle": BaseActuatorMetaCfg(),
        "right_hip_yaw": BaseActuatorMetaCfg(),
        "right_hip_roll": BaseActuatorMetaCfg(),
        "right_hip_pitch": BaseActuatorMetaCfg(),
        "right_knee": BaseActuatorMetaCfg(),
        "right_ankle": BaseActuatorMetaCfg(),
        "torso": BaseActuatorMetaCfg(),
        "left_shoulder_pitch": BaseActuatorMetaCfg(),
        "left_shoulder_roll": BaseActuatorMetaCfg(),
        "left_shoulder_yaw": BaseActuatorMetaCfg(),
        "left_elbow": BaseActuatorMetaCfg(),
        "left_hand": BaseActuatorMetaCfg(),
        "right_shoulder_pitch": BaseActuatorMetaCfg(),
        "right_shoulder_roll": BaseActuatorMetaCfg(),
        "right_shoulder_yaw": BaseActuatorMetaCfg(),
        "right_elbow": BaseActuatorMetaCfg(),
        "right_hand": BaseActuatorMetaCfg(),
        "L_index_proximal": BaseActuatorMetaCfg(),
        "L_index_intermediate": BaseActuatorMetaCfg(),
        "L_middle_proximal": BaseActuatorMetaCfg(),
        "L_middle_intermediate": BaseActuatorMetaCfg(),
        "L_ring_proximal": BaseActuatorMetaCfg(),
        "L_ring_intermediate": BaseActuatorMetaCfg(),
        "L_pinky_proximal": BaseActuatorMetaCfg(),
        "L_pinky_intermediate": BaseActuatorMetaCfg(),
        "L_thumb_proximal_yaw": BaseActuatorMetaCfg(),
        "L_thumb_proximal_pitch": BaseActuatorMetaCfg(),
        "L_thumb_intermediate": BaseActuatorMetaCfg(),
        "L_thumb_distal": BaseActuatorMetaCfg(),
        "R_index_proximal": BaseActuatorMetaCfg(),
        "R_index_intermediate": BaseActuatorMetaCfg(),
        "R_middle_proximal": BaseActuatorMetaCfg(),
        "R_middle_intermediate": BaseActuatorMetaCfg(),
        "R_ring_proximal": BaseActuatorMetaCfg(),
        "R_ring_intermediate": BaseActuatorMetaCfg(),
        "R_pinky_proximal": BaseActuatorMetaCfg(),
        "R_pinky_intermediate": BaseActuatorMetaCfg(),
        "R_thumb_proximal_yaw": BaseActuatorMetaCfg(),
        "R_thumb_proximal_pitch": BaseActuatorMetaCfg(),
        "R_thumb_intermediate": BaseActuatorMetaCfg(),
        "R_thumb_distal": BaseActuatorMetaCfg(),
    }
    joint_limits: dict[str, tuple[float, float]] = {
        "left_hip_yaw": (-0.43, 0.43),
        "left_hip_roll": (-0.43, 0.43),
        "left_hip_pitch": (-3.14, 2.53),
        "left_knee": (-0.26, 2.05),
        "left_ankle": (-0.87, 0.52),
        "right_hip_yaw": (-0.43, 0.43),
        "right_hip_roll": (-0.43, 0.43),
        "right_hip_pitch": (-3.14, 2.53),
        "right_knee": (-0.26, 2.05),
        "right_ankle": (-0.87, 0.52),
        "torso": (-2.35, 2.35),
        "left_shoulder_pitch": (-2.87, 2.87),
        "left_shoulder_roll": (-0.34, 3.11),
        "left_shoulder_yaw": (-1.3, 4.45),
        "left_elbow": (-1.25, 2.61),
        "left_hand": (-6, 6),
        "right_shoulder_pitch": (-2.87, 2.87),
        "right_shoulder_roll": (-3.11, 0.34),
        "right_shoulder_yaw": (-4.45, 1.3),
        "right_elbow": (-1.25, 2.61),
        "right_hand": (-6, 6),
        "L_index_proximal": (-1, 1),
        "L_index_intermediate": (-1, 1),
        "L_middle_proximal": (-1, 1),
        "L_middle_intermediate": (-1, 1),
        "L_ring_proximal": (-1, 1),
        "L_ring_intermediate": (-1, 1),
        "L_pinky_proximal": (-1, 1),
        "L_pinky_intermediate": (-1, 1),
        "L_thumb_proximal_yaw": (-1, 1),
        "L_thumb_proximal_pitch": (-1, 1),
        "L_thumb_intermediate": (-1, 1),
        "L_thumb_distal": (-1, 1),
        "R_index_proximal": (-1, 1),
        "R_index_intermediate": (-1, 1),
        "R_middle_proximal": (-1, 1),
        "R_middle_intermediate": (-1, 1),
        "R_ring_proximal": (-1, 1),
        "R_ring_intermediate": (-1, 1),
        "R_pinky_proximal": (-1, 1),
        "R_pinky_intermediate": (-1, 1),
        "R_thumb_proximal_yaw": (-1, 1),
        "R_thumb_proximal_pitch": (-1, 1),
        "R_thumb_intermediate": (-1, 1),
        "R_thumb_distal": (-1, 1),
    }



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, PhysicStateType, TaskType
from metasim.utils import configclass


@configclass
class OpensdorRotLighterForthLighter428MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/b640aa4990b64db9ab3d868c6f49820e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/b640aa4990b64db9ab3d868c6f49820e_lighter_forth/20240826-213920_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter429MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/b640aa4990b64db9ab3d868c6f49820e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/b640aa4990b64db9ab3d868c6f49820e_lighter_forth/20240826-211552_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter430MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/b640aa4990b64db9ab3d868c6f49820e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/b640aa4990b64db9ab3d868c6f49820e_lighter_forth/20240826-215210_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter431MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/b640aa4990b64db9ab3d868c6f49820e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/b640aa4990b64db9ab3d868c6f49820e_lighter_forth/20240826-214058_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter432MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/b640aa4990b64db9ab3d868c6f49820e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/b640aa4990b64db9ab3d868c6f49820e_lighter_forth/20240826-212547_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter433MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/b640aa4990b64db9ab3d868c6f49820e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/b640aa4990b64db9ab3d868c6f49820e_lighter_forth/20240826-213250_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter490MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/d9675ab05c39447baf27e19ea07d484e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/d9675ab05c39447baf27e19ea07d484e_lighter_forth/20240826-211406_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter491MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/d9675ab05c39447baf27e19ea07d484e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/d9675ab05c39447baf27e19ea07d484e_lighter_forth/20240826-215856_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter492MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/d9675ab05c39447baf27e19ea07d484e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/d9675ab05c39447baf27e19ea07d484e_lighter_forth/20240826-213520_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter493MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/d9675ab05c39447baf27e19ea07d484e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/d9675ab05c39447baf27e19ea07d484e_lighter_forth/20240826-213942_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter494MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/d9675ab05c39447baf27e19ea07d484e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/d9675ab05c39447baf27e19ea07d484e_lighter_forth/20240826-213812_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotLighterForthLighter495MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="lighter",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/d9675ab05c39447baf27e19ea07d484e/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/d9675ab05c39447baf27e19ea07d484e_lighter_forth/20240826-220537_no_interaction/trajectory-unified_wo_traj_v2.pkl"



import torch


def get_cam_params(
    cam_pos: torch.Tensor,
    cam_look_at: torch.Tensor,
    width=640,
    height=480,
    focal_length=24,
    horizontal_aperture=20.955,
    vertical_aperture=None,
):
    """Copy from old version"""
    if vertical_aperture is None:
        vertical_aperture = horizontal_aperture * height / width

    device = cam_pos.device
    num_envs = len(cam_pos)
    cam_front = cam_look_at - cam_pos
    cam_right = torch.cross(cam_front, torch.tensor([[0.0, 0.0, 1.0]], device=device), dim=1)
    cam_up = torch.cross(cam_right, cam_front)

    cam_right = cam_right / (torch.norm(cam_right, dim=-1, keepdim=True) + 1e-12)
    cam_front = cam_front / (torch.norm(cam_front, dim=-1, keepdim=True) + 1e-12)
    cam_up = cam_up / (torch.norm(cam_up, dim=-1, keepdim=True) + 1e-12)

    # Camera convention difference between ROS and Isaac Sim
    R = torch.stack([cam_right, -cam_up, cam_front], dim=1)  # .transpose(-1, -2)
    t = -torch.bmm(R, cam_pos.unsqueeze(-1)).squeeze()
    extrinsics = torch.eye(4, device=device).unsqueeze(0).tile([num_envs, 1, 1])
    extrinsics[:, :3, :3] = R
    extrinsics[:, :3, 3] = t

    fx = width * focal_length / horizontal_aperture
    fy = height * focal_length / vertical_aperture
    cx = width * 0.5
    cy = height * 0.5

    intrinsics = (
        torch.tensor([[fx, 0.0, cx], [0.0, fy, cy], [0.0, 0.0, 1.0]], device=device).unsqueeze(0).tile([num_envs, 1, 1])
    )

    return extrinsics, intrinsics



from __future__ import annotations

import argparse
import json
import os
import time

import isaacgym  # noqa: F401
import numpy as np
import rootutils
import torch
from curobo.types.math import Pose
from loguru import logger as log
from pytorch3d import transforms
from rich.logging import RichHandler

rootutils.setup_root(__file__, pythonpath=True)
log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--num_envs", type=int, default=1)
    parser.add_argument("--sim", type=str, default="isaaclab", choices=["isaaclab", "isaacgym", "mujoco"])
    parser.add_argument("--task", type=str, default="PickCube")
    parser.add_argument(
        "--robot",
        type=str,
        default="franka",
        choices=[
            "franka",
            "ur5e_2f85",
            "sawyer",
            "franka_with_gripper_extension",
            "h1_2_without_hand",
            "h1",
            "h1_simple_hand",
            "sawyer_mujoco",
            "fetch",
        ],
    )
    parser.add_argument("--add_table", action="store_true")
    parser.add_argument(
        "--joints", default=None, nargs="+", type=str, help="Joints to randomize, if None, randomize all joints"
    )
    args = parser.parse_args()
    return args


def get_gapartnet_anno(anno_path):
    """
    Get gapartnet annotation
    """
    info = {}
    # load object annotation
    anno = json.loads(open(anno_path).read())
    num_link_anno = len(anno)
    gapart_raw_valid_anno = []
    for link_i in range(num_link_anno):
        anno_i = anno[link_i]
        if anno_i["is_gapart"]:
            gapart_raw_valid_anno.append(anno_i)
    info["gapart_cates"] = [anno_i["category"] for anno_i in gapart_raw_valid_anno]
    info["gapart_init_bboxes"] = np.array([np.asarray(anno_i["bbox"]) for anno_i in gapart_raw_valid_anno])
    info["gapart_link_names"] = [anno_i["link_name"] for anno_i in gapart_raw_valid_anno]
    return info


def control_to_pose(
    env,
    num_envs,
    robot,
    ee_pos_target,
    rotation,
    gripper_widths,
    robot_ik,
    curobo_n_dof,
    ee_n_dof,
    seed_config,
    steps=3,
):
    pass
    # Solve IK
    result = robot_ik.solve_batch(Pose(ee_pos_target, rotation=rotation), seed_config=seed_config)

    q = torch.zeros((num_envs, robot.num_joints), device="cuda:0")
    ik_succ = result.success.squeeze(1)
    q[ik_succ, :curobo_n_dof] = result.solution[ik_succ, 0].clone()
    q[:, -ee_n_dof:] = gripper_widths

    actions = [{"dof_pos_target": dict(zip(robot.actuators.keys(), q[i_env].tolist()))} for i_env in range(num_envs)]

    for i_step in range(steps):
        env.step(actions)
        env.render()
    return actions


def main():
    args = parse_args()
    num_envs: int = args.num_envs

    # specificly for isaacgym
    if args.sim == "isaacgym":
        from isaacgym import gymapi, gymtorch, gymutil  # noqa: F401

    ## Import put here to support isaacgym

    import torch

    from metasim.cfg.scenario import ScenarioMetaCfg
    from metasim.constants import SimType
    from metasim.utils.demo_util import get_traj
    from metasim.utils.kinematics_utils import get_curobo_models
    from metasim.utils.setup_util import get_robot, get_sim_env_class

    scenario = ScenarioMetaCfg(task=args.task, robot=args.robot, try_add_table=args.add_table, sim=args.sim)

    robot = get_robot(args.robot)
    *_, robot_ik = get_curobo_models(robot)
    curobo_n_dof = len(robot_ik.robot_config.cspace.joint_names)
    ee_n_dof = len(robot.gripper_release_q)

    log.info(f"Using simulator: {args.sim}")
    env_class = get_sim_env_class(SimType(args.sim))
    env = env_class(scenario, num_envs)

    ## Main
    tic = time.time()
    assert os.path.exists(scenario.task.traj_filepath), (
        f"Trajectory file: {scenario.task.traj_filepath} does not exist."
    )
    init_states, all_actions, all_states = get_traj(scenario.task, scenario.robot, env.handler)
    toc = time.time()
    log.trace(f"Time to load data: {toc - tic:.2f}s")
    ## Reset before first step
    tic = time.time()
    obs, extras = env.reset(states=init_states[:num_envs])
    toc = time.time()
    log.trace(f"Time to reset: {toc - tic:.2f}s")

    # Generate random actions
    urdf_path = scenario.task.objects[0].urdf_path
    anno_path = urdf_path.replace("mobility_annotation_gapartnet.urdf", "link_annotation_gapartnet.json")
    gapartnet_anno = get_gapartnet_anno(anno_path)

    ee_rot_target = torch.rand((num_envs, 3), device="cuda:0") * torch.pi
    ee_quat_target = transforms.matrix_to_quaternion(transforms.euler_angles_to_matrix(ee_rot_target, "XYZ"))

    # Compute targets
    curr_robot_q = obs["joint_qpos"].cuda()
    robot_root_state = obs["robot_root_state"].cuda()
    robot_pos, robot_quat = robot_root_state[:, 0:3], robot_root_state[:, 3:7]
    seed_config = curr_robot_q[:, :curobo_n_dof].unsqueeze(1).tile([1, robot_ik._num_seeds, 1])

    bbox_id = -1
    all_bbox_now = gapartnet_anno["gapart_init_bboxes"] * scenario.task.objects[0].scale + np.array(
        init_states[0]["cabinet"]["pos"]
    )
    # get the part bbox and calculate the handle direction
    all_bbox_now = torch.tensor(all_bbox_now, dtype=torch.float32).to("cuda:0").reshape(-1, 8, 3)
    all_bbox_center_front_face = torch.mean(all_bbox_now[:, 0:4, :], dim=1)
    handle_out = all_bbox_now[:, 0, :] - all_bbox_now[:, 4, :]
    handle_out /= torch.norm(handle_out, dim=1, keepdim=True)
    handle_long = all_bbox_now[:, 0, :] - all_bbox_now[:, 1, :]
    handle_long /= torch.norm(handle_long, dim=1, keepdim=True)
    handle_short = all_bbox_now[:, 0, :] - all_bbox_now[:, 3, :]
    handle_short /= torch.norm(handle_short, dim=1, keepdim=True)

    init_position = all_bbox_center_front_face[bbox_id]
    handle_out_ = handle_out[bbox_id]

    gripper_widths = torch.tensor(robot.gripper_release_q).to("cuda:0")

    rotation_transform_for_franka = torch.tensor(
        [
            [0.0, 0.0, 1.0],
            [0.0, -1.0, 0.0],
            [1.0, 0.0, 0.0],
        ],
        device="cuda:0",
    )
    rotation_target = torch.stack(
        [
            -handle_out[bbox_id],
            -handle_short[bbox_id],
            handle_long[bbox_id],
        ],
        dim=0,
    )
    rotation = rotation_target @ rotation_transform_for_franka
    log.info("move to pre-grasp position")
    for i in range(30):
        control_to_pose(
            env,
            num_envs,
            robot,
            init_position + 0.2 * handle_out_,
            rotation,
            gripper_widths,
            robot_ik,
            curobo_n_dof,
            ee_n_dof,
            seed_config,
        )
    # move the object to the grasp position
    log.info("move to grasp position")
    for i in range(20):
        actions = control_to_pose(
            env,
            num_envs,
            robot,
            init_position + (0.1 - 0.015 * i) * handle_out_,
            rotation,
            gripper_widths,
            robot_ik,
            curobo_n_dof,
            ee_n_dof,
            seed_config,
            steps=3,
        )

    # close the gripper
    log.info("close the gripper")

    gripper_widths = torch.tensor(robot.gripper_release_q)
    gripper_widths[:] = 0.0
    actions[0]["dof_pos_target"]["panda_finger_joint1"] = 0.0
    actions[0]["dof_pos_target"]["panda_finger_joint2"] = 0.0
    for i in range(10):
        env.step(actions)
        env.render()

    # move the object to the lift position]
    log.info("move the object to the lift position")
    for i in range(30):
        control_to_pose(
            env,
            num_envs,
            robot,
            init_position + (-0.2 + i * 0.01) * handle_out_,
            rotation,
            gripper_widths,
            robot_ik,
            curobo_n_dof,
            ee_n_dof,
            seed_config,
        )

    log.info("done")


if __name__ == "__main__":
    main()



## TODO: This code need to be checked carefully when upgrading to new IsaacLab version

from typing import Sequence

import torch
from omni.isaac.lab.envs.common import VecEnvObs, VecEnvStepReturn
from omni.isaac.lab.envs.direct_rl_env import DirectRLEnv


## This env class is almost same as DirectRLEnv
class CustomDirectRLEnv(DirectRLEnv):
    def reset(
        self,
        env_ids: Sequence[int] | None = None,  # ! new argument
        seed: int | None = None,
        options=None,
    ) -> tuple[VecEnvObs, dict]:
        """
        Compared to `DirectRLEnv.reset()`, this function support resetting specific environments
        """
        # set the seed
        if seed is not None:
            self.seed(seed)

        # reset state of scene
        if env_ids is None:
            env_ids = torch.arange(self.num_envs, dtype=torch.int64, device=self.device)
        elif isinstance(env_ids, list):
            env_ids = torch.tensor(env_ids, dtype=torch.int64, device=self.device)
        self._reset_idx(env_ids)

        # if sensors are added to the scene, make sure we render to reflect changes in reset
        if self.sim.has_rtx_sensors() and self.cfg.rerender_on_reset:
            self.sim.render()

        # return observations
        return self._get_observations(), self.extras

    def step(self, action: torch.Tensor) -> VecEnvStepReturn:
        """
        Compared to DirectRLEnv.step(), this function won't automatically reset environments that is successful or timed-out.
        """
        action = action.to(self.device)
        # add action noise
        if self.cfg.action_noise_model:
            action = self._action_noise_model.apply(action)

        # process actions
        self._pre_physics_step(action)

        # check if we need to do rendering within the physics loop
        # note: checked here once to avoid multiple checks within the loop
        is_rendering = self.sim.has_gui() or self.sim.has_rtx_sensors()

        # perform physics stepping
        for _ in range(self.cfg.decimation):
            self._sim_step_counter += 1
            # set actions into buffers
            self._apply_action()
            # set actions into simulator
            self.scene.write_data_to_sim()
            # simulate
            self.sim.step(render=False)
            # render between steps only if the GUI or an RTX sensor needs it
            # note: we assume the render interval to be the shortest accepted rendering interval.
            #    If a camera needs rendering at a faster frequency, this will lead to unexpected behavior.
            if self._sim_step_counter % self.cfg.sim.render_interval == 0 and is_rendering:
                self.sim.render()
            # update buffers at sim dt
            self.scene.update(dt=self.physics_dt)

        # post-step:
        # -- update env counters (used for curriculum generation)
        self.episode_length_buf += 1  # step in current episode (per env)
        self.common_step_counter += 1  # total step (common for all envs)

        self.reset_terminated[:], self.reset_time_outs[:] = self._get_dones()
        self.reset_buf = self.reset_terminated | self.reset_time_outs
        self.reward_buf = self._get_rewards()

        # ! Below is the original code from DirectRLEnv.step()
        # # -- reset envs that terminated/timed-out and log the episode information
        # reset_env_ids = self.reset_buf.nonzero(as_tuple=False).squeeze(-1)
        # if len(reset_env_ids) > 0:
        #     self._reset_idx(reset_env_ids)
        #     # if sensors are added to the scene, make sure we render to reflect changes in reset
        #     if self.sim.has_rtx_sensors() and self.cfg.rerender_on_reset:
        #         self.sim.render()
        # ! End of original code

        # post-step: step interval event
        if self.cfg.events:
            if "interval" in self.event_manager.available_modes:
                self.event_manager.apply(mode="interval", dt=self.step_dt)

        # update observations
        self.obs_buf = None  # XXX: modified to speed up

        # add observation noise
        # note: we apply no noise to the state space (since it is used for critic networks)
        if self.cfg.observation_noise_model:
            self.obs_buf["policy"] = self._observation_noise_model.apply(self.obs_buf["policy"])

        # return observations, rewards, resets and extras
        return self.obs_buf, self.reward_buf, self.reset_terminated, self.reset_time_outs, self.extras

    def get_observations(self):
        return self._get_observations()

    def get_dones(self):
        return self._get_dones()
